@@ -9,7 +9,7 @@ from . import path_random
 from .contract import contract, contract_path, contract_expression
 from .parser import get_symbol
 from .sharing import shared_intermediates
-from .paths import BranchBound
+from .paths import BranchBound, DynamicProgramming
 from .path_random import RandomGreedy
 
 # Handle versioneer
@@ -19,5 +19,5 @@ __version__ = versions['version']
 __git_revision__ = versions['full-revisionid']
 del get_versions, versions
 
-
 paths.register_path_fn('random-greedy', path_random.random_greedy)
+paths.register_path_fn('random-greedy-128', path_random.random_greedy_128)
